import json
from uuid import UUID, uuid4

from eventstoredb import Client
from eventstoredb.events import JsonEvent, JsonRecordedEvent, ReadEvent, RecordedEvent
from eventstoredb.streams.read import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    ReadAllOptions,
    ReadDirection,
    StreamNameFilter,
)
from eventstoredb.streams.types import AllPosition, StreamPosition


async def test_read_all(eventstoredb_client: Client) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[
            JsonEvent(
                type="test_read_all_Test1",
                data=json.dumps({"some": "data"}).encode(),
            )
        ],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[
            JsonEvent(
                type="test_read_all_Test2",
                data=json.dumps({"some": "data"}).encode(),
            )
        ],
    )

    it = eventstoredb_client.read_all()
    # read all events and filter for events with types starting with 'test_read_all_'
    # the filtering is done because the eventstore is shared across the pytest-session
    events = [
        e async for e in it if e.event and e.event.type.startswith("test_read_all_")
    ]

    assert len(events) == 2

    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, JsonRecordedEvent)
    assert isinstance(event.id, UUID)
    assert event.stream_name == stream_name_1
    assert event.type == "test_read_all_Test1"
    assert event.revision == 0
    assert event.data == json.dumps({"some": "data"}).encode()
    assert event.metadata is None

    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, JsonRecordedEvent)
    assert isinstance(event.id, UUID)
    assert event.stream_name == stream_name_2
    assert event.type == "test_read_all_Test2"
    assert event.revision == 0
    assert event.data == json.dumps({"some": "data"}).encode()
    assert event.metadata is None


async def test_read_all_from_start(eventstoredb_client: Client) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="test_read_all_from_start_Test1")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="test_read_all_from_start_Test2")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(from_position=StreamPosition.START)
    )
    # read all events and filter by event-type containing the right prefix
    # the filtering is done because the eventstore is shared across the pytest-session
    events = [
        e
        async for e in it
        if e.event and e.event.type.startswith("test_read_all_from_start_")
    ]

    assert len(events) == 2

    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_1
    assert event.type == "test_read_all_from_start_Test1"

    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_2
    assert event.type == "test_read_all_from_start_Test2"


async def test_read_all_from_position(eventstoredb_client: Client) -> None:
    stream_name = f"Test-{uuid4()}"
    # write some events onto stream
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Junk"), JsonEvent(type="Junk")],
    )
    # write one event to get current stream-position
    marker = await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Marker")],
    )
    assert marker.position is not None

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Test")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(
            from_position=AllPosition(
                commit=marker.position.commit,
                prepare=marker.position.prepare,
            )
        )
    )
    # read all events and filter out all system events (starting with '$')
    events = [e async for e in it if e.event and not e.event.type.startswith("$")]

    # for some odd reason read_all includes the event _at_ position
    # which seems to be different than subscribe_to_all
    assert len(events) == 2

    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "Marker"

    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "Test"


async def test_read_all_backwards_from_end(eventstoredb_client: Client) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="test_read_all_backwards_from_end_Test1")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="test_read_all_backwards_from_end_Test2")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(
            from_position=StreamPosition.END,
            direction=ReadDirection.BACKWARDS,
        )
    )
    # read all events and filter by event-type containing the right prefix
    # the filtering is done because the eventstore is shared across the pytest-session
    events = [
        e
        async for e in it
        if e.event and e.event.type.startswith("test_read_all_backwards_from_end_")
    ]

    assert len(events) == 2

    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_2
    assert event.type == "test_read_all_backwards_from_end_Test2"

    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_1
    assert event.type == "test_read_all_backwards_from_end_Test1"


async def test_read_all_backwards_from_position(eventstoredb_client: Client) -> None:
    stream_name = f"Test-{uuid4()}"
    # write some events onto stream
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Test1"), JsonEvent(type="Test2")],
    )
    # write one event to get current stream-position
    marker = await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Marker")],
    )
    assert marker.position is not None
    # write some events as junk
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Junk"), JsonEvent(type="Junk")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(
            from_position=AllPosition(
                commit=marker.position.commit,
                prepare=marker.position.prepare,
            ),
            direction=ReadDirection.BACKWARDS,
            max_count=2,
        )
    )
    # read all events and filter out all system events (starting with '$')
    events = [e async for e in it if e.event and not e.event.type.startswith("$")]
    for e in events:
        print(e)

    assert len(events) == 2

    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "Test2"

    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "Test1"


async def test_read_all_exclude_system_events(eventstoredb_client: Client) -> None:
    stream_name = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="test_read_all_exclude_system_events_Test")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(filter=ExcludeSystemEventsFilter())
    )

    events = [
        e
        async for e in it
        if e.event
        and (
            e.event.type.startswith("$")
            or e.event.type.startswith("test_read_all_exclude_system_events_")
        )
    ]

    assert len(events) == 1
    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "test_read_all_exclude_system_events_Test"


async def test_read_all_filter_by_event_type_regex(eventstoredb_client: Client) -> None:
    stream_name = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_read_all_filter_by_event_type_regex_Test1"),
            JsonEvent(type="test_read_all_filter_by_event_type_regex_Test2"),
        ],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(
            filter=EventTypeFilter(
                regex="test_read_all_filter_by_event_type_regex_Test1"
            )
        )
    )

    events = [e async for e in it]

    assert len(events) == 1
    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "test_read_all_filter_by_event_type_regex_Test1"


async def test_read_all_filter_by_event_type_prefix(
    eventstoredb_client: Client,
) -> None:
    stream_name = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_read_all_filter_by_event_type_prefix_Test1"),
            JsonEvent(type="test_read_all_filter_by_event_type_prefix_Test2"),
            JsonEvent(type="something_else"),
        ],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(
            filter=EventTypeFilter(
                prefix=["test_read_all_filter_by_event_type_prefix_"]
            )
        )
    )

    events = [e async for e in it]

    assert len(events) == 2
    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "test_read_all_filter_by_event_type_prefix_Test1"
    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert event.type == "test_read_all_filter_by_event_type_prefix_Test2"


async def test_read_all_filter_by_stream_name_regex(
    eventstoredb_client: Client,
) -> None:
    stream_name_1 = f"Test-{uuid4()}"
    stream_name_2 = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(filter=StreamNameFilter(regex=stream_name_1))
    )

    events = [e async for e in it]

    assert len(events) == 1
    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_1
    assert event.type == "Test"


async def test_read_all_filter_by_stream_name_prefix(
    eventstoredb_client: Client,
) -> None:
    stream_name_1 = f"FilterByStreamName1-{uuid4()}"
    stream_name_2 = f"FilterByStreamName2-{uuid4()}"
    stream_name_3 = f"Junk-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_3,
        events=[JsonEvent(type="Test")],
    )

    it = eventstoredb_client.read_all(
        options=ReadAllOptions(filter=StreamNameFilter(prefix=["FilterByStreamName"]))
    )

    events = [e async for e in it]

    assert len(events) == 2
    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_1
    assert event.type == "Test"
    assert isinstance(events[1], ReadEvent)
    event = events[1].event
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name_2
    assert event.type == "Test"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: streams.proto
# plugin: python-betterproto
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from ....google import rpc as ___google_rpc__
from ... import client as __client__

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class ReadReqOptionsReadDirection(betterproto.Enum):
    Forwards = 0
    Backwards = 1


@dataclass(eq=False, repr=False)
class ReadReq(betterproto.Message):
    options: "ReadReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ReadReqOptions(betterproto.Message):
    stream: "ReadReqOptionsStreamOptions" = betterproto.message_field(
        1, group="stream_option"
    )
    all: "ReadReqOptionsAllOptions" = betterproto.message_field(
        2, group="stream_option"
    )
    read_direction: "ReadReqOptionsReadDirection" = betterproto.enum_field(3)
    resolve_links: bool = betterproto.bool_field(4)
    count: int = betterproto.uint64_field(5, group="count_option")
    subscription: "ReadReqOptionsSubscriptionOptions" = betterproto.message_field(
        6, group="count_option"
    )
    filter: "ReadReqOptionsFilterOptions" = betterproto.message_field(
        7, group="filter_option"
    )
    no_filter: "__client__.Empty" = betterproto.message_field(8, group="filter_option")
    uuid_option: "ReadReqOptionsUuidOption" = betterproto.message_field(9)
    control_option: "ReadReqOptionsControlOption" = betterproto.message_field(10)


@dataclass(eq=False, repr=False)
class ReadReqOptionsStreamOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    revision: int = betterproto.uint64_field(2, group="revision_option")
    start: "__client__.Empty" = betterproto.message_field(3, group="revision_option")
    end: "__client__.Empty" = betterproto.message_field(4, group="revision_option")


@dataclass(eq=False, repr=False)
class ReadReqOptionsAllOptions(betterproto.Message):
    position: "ReadReqOptionsPosition" = betterproto.message_field(
        1, group="all_option"
    )
    start: "__client__.Empty" = betterproto.message_field(2, group="all_option")
    end: "__client__.Empty" = betterproto.message_field(3, group="all_option")


@dataclass(eq=False, repr=False)
class ReadReqOptionsSubscriptionOptions(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ReadReqOptionsPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ReadReqOptionsFilterOptions(betterproto.Message):
    stream_identifier: "ReadReqOptionsFilterOptionsExpression" = (
        betterproto.message_field(1, group="filter")
    )
    event_type: "ReadReqOptionsFilterOptionsExpression" = betterproto.message_field(
        2, group="filter"
    )
    max: int = betterproto.uint32_field(3, group="window")
    count: "__client__.Empty" = betterproto.message_field(4, group="window")
    checkpoint_interval_multiplier: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class ReadReqOptionsFilterOptionsExpression(betterproto.Message):
    regex: str = betterproto.string_field(1)
    prefix: List[str] = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ReadReqOptionsUuidOption(betterproto.Message):
    structured: "__client__.Empty" = betterproto.message_field(1, group="content")
    string: "__client__.Empty" = betterproto.message_field(2, group="content")


@dataclass(eq=False, repr=False)
class ReadReqOptionsControlOption(betterproto.Message):
    compatibility: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ReadResp(betterproto.Message):
    event: "ReadRespReadEvent" = betterproto.message_field(1, group="content")
    confirmation: "ReadRespSubscriptionConfirmation" = betterproto.message_field(
        2, group="content"
    )
    checkpoint: "ReadRespCheckpoint" = betterproto.message_field(3, group="content")
    stream_not_found: "ReadRespStreamNotFound" = betterproto.message_field(
        4, group="content"
    )
    first_stream_position: int = betterproto.uint64_field(5, group="content")
    last_stream_position: int = betterproto.uint64_field(6, group="content")
    last_all_stream_position: "__client__.AllStreamPosition" = (
        betterproto.message_field(7, group="content")
    )


@dataclass(eq=False, repr=False)
class ReadRespReadEvent(betterproto.Message):
    event: "ReadRespReadEventRecordedEvent" = betterproto.message_field(1)
    link: "ReadRespReadEventRecordedEvent" = betterproto.message_field(2)
    commit_position: int = betterproto.uint64_field(3, group="position")
    no_position: "__client__.Empty" = betterproto.message_field(4, group="position")


@dataclass(eq=False, repr=False)
class ReadRespReadEventRecordedEvent(betterproto.Message):
    id: "__client__.Uuid" = betterproto.message_field(1)
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(2)
    stream_revision: int = betterproto.uint64_field(3)
    prepare_position: int = betterproto.uint64_field(4)
    commit_position: int = betterproto.uint64_field(5)
    metadata: Dict[str, str] = betterproto.map_field(
        6, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    custom_metadata: bytes = betterproto.bytes_field(7)
    data: bytes = betterproto.bytes_field(8)


@dataclass(eq=False, repr=False)
class ReadRespSubscriptionConfirmation(betterproto.Message):
    subscription_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ReadRespCheckpoint(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ReadRespStreamNotFound(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class AppendReq(betterproto.Message):
    options: "AppendReqOptions" = betterproto.message_field(1, group="content")
    proposed_message: "AppendReqProposedMessage" = betterproto.message_field(
        2, group="content"
    )


@dataclass(eq=False, repr=False)
class AppendReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    revision: int = betterproto.uint64_field(2, group="expected_stream_revision")
    no_stream: "__client__.Empty" = betterproto.message_field(
        3, group="expected_stream_revision"
    )
    any: "__client__.Empty" = betterproto.message_field(
        4, group="expected_stream_revision"
    )
    stream_exists: "__client__.Empty" = betterproto.message_field(
        5, group="expected_stream_revision"
    )


@dataclass(eq=False, repr=False)
class AppendReqProposedMessage(betterproto.Message):
    id: "__client__.Uuid" = betterproto.message_field(1)
    metadata: Dict[str, str] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    custom_metadata: bytes = betterproto.bytes_field(3)
    data: bytes = betterproto.bytes_field(4)


@dataclass(eq=False, repr=False)
class AppendResp(betterproto.Message):
    success: "AppendRespSuccess" = betterproto.message_field(1, group="result")
    wrong_expected_version: "AppendRespWrongExpectedVersion" = (
        betterproto.message_field(2, group="result")
    )


@dataclass(eq=False, repr=False)
class AppendRespPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class AppendRespSuccess(betterproto.Message):
    current_revision: int = betterproto.uint64_field(1, group="current_revision_option")
    no_stream: "__client__.Empty" = betterproto.message_field(
        2, group="current_revision_option"
    )
    position: "AppendRespPosition" = betterproto.message_field(
        3, group="position_option"
    )
    no_position: "__client__.Empty" = betterproto.message_field(
        4, group="position_option"
    )


@dataclass(eq=False, repr=False)
class AppendRespWrongExpectedVersion(betterproto.Message):
    current_revision_20_6_0: int = betterproto.uint64_field(
        1, group="current_revision_option_20_6_0"
    )
    no_stream_20_6_0: "__client__.Empty" = betterproto.message_field(
        2, group="current_revision_option_20_6_0"
    )
    expected_revision_20_6_0: int = betterproto.uint64_field(
        3, group="expected_revision_option_20_6_0"
    )
    any_20_6_0: "__client__.Empty" = betterproto.message_field(
        4, group="expected_revision_option_20_6_0"
    )
    stream_exists_20_6_0: "__client__.Empty" = betterproto.message_field(
        5, group="expected_revision_option_20_6_0"
    )
    current_revision: int = betterproto.uint64_field(6, group="current_revision_option")
    current_no_stream: "__client__.Empty" = betterproto.message_field(
        7, group="current_revision_option"
    )
    expected_revision: int = betterproto.uint64_field(
        8, group="expected_revision_option"
    )
    expected_any: "__client__.Empty" = betterproto.message_field(
        9, group="expected_revision_option"
    )
    expected_stream_exists: "__client__.Empty" = betterproto.message_field(
        10, group="expected_revision_option"
    )
    expected_no_stream: "__client__.Empty" = betterproto.message_field(
        11, group="expected_revision_option"
    )


@dataclass(eq=False, repr=False)
class BatchAppendReq(betterproto.Message):
    correlation_id: "__client__.Uuid" = betterproto.message_field(1)
    options: "BatchAppendReqOptions" = betterproto.message_field(2)
    proposed_messages: List[
        "BatchAppendReqProposedMessage"
    ] = betterproto.message_field(3)
    is_final: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class BatchAppendReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    stream_position: int = betterproto.uint64_field(2, group="expected_stream_position")
    no_stream: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        3, group="expected_stream_position"
    )
    any: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        4, group="expected_stream_position"
    )
    stream_exists: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        5, group="expected_stream_position"
    )
    deadline_21_10_0: datetime = betterproto.message_field(6, group="deadline_option")
    deadline: timedelta = betterproto.message_field(7, group="deadline_option")


@dataclass(eq=False, repr=False)
class BatchAppendReqProposedMessage(betterproto.Message):
    id: "__client__.Uuid" = betterproto.message_field(1)
    metadata: Dict[str, str] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    custom_metadata: bytes = betterproto.bytes_field(3)
    data: bytes = betterproto.bytes_field(4)


@dataclass(eq=False, repr=False)
class BatchAppendResp(betterproto.Message):
    correlation_id: "__client__.Uuid" = betterproto.message_field(1)
    error: "___google_rpc__.Status" = betterproto.message_field(2, group="result")
    success: "BatchAppendRespSuccess" = betterproto.message_field(3, group="result")
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(4)
    stream_position: int = betterproto.uint64_field(5, group="expected_stream_position")
    no_stream: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        6, group="expected_stream_position"
    )
    any: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        7, group="expected_stream_position"
    )
    stream_exists: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        8, group="expected_stream_position"
    )


@dataclass(eq=False, repr=False)
class BatchAppendRespSuccess(betterproto.Message):
    current_revision: int = betterproto.uint64_field(1, group="current_revision_option")
    no_stream: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        2, group="current_revision_option"
    )
    position: "__client__.AllStreamPosition" = betterproto.message_field(
        3, group="position_option"
    )
    no_position: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        4, group="position_option"
    )


@dataclass(eq=False, repr=False)
class DeleteReq(betterproto.Message):
    options: "DeleteReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    revision: int = betterproto.uint64_field(2, group="expected_stream_revision")
    no_stream: "__client__.Empty" = betterproto.message_field(
        3, group="expected_stream_revision"
    )
    any: "__client__.Empty" = betterproto.message_field(
        4, group="expected_stream_revision"
    )
    stream_exists: "__client__.Empty" = betterproto.message_field(
        5, group="expected_stream_revision"
    )


@dataclass(eq=False, repr=False)
class DeleteResp(betterproto.Message):
    position: "DeleteRespPosition" = betterproto.message_field(
        1, group="position_option"
    )
    no_position: "__client__.Empty" = betterproto.message_field(
        2, group="position_option"
    )


@dataclass(eq=False, repr=False)
class DeleteRespPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class TombstoneReq(betterproto.Message):
    options: "TombstoneReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class TombstoneReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    revision: int = betterproto.uint64_field(2, group="expected_stream_revision")
    no_stream: "__client__.Empty" = betterproto.message_field(
        3, group="expected_stream_revision"
    )
    any: "__client__.Empty" = betterproto.message_field(
        4, group="expected_stream_revision"
    )
    stream_exists: "__client__.Empty" = betterproto.message_field(
        5, group="expected_stream_revision"
    )


@dataclass(eq=False, repr=False)
class TombstoneResp(betterproto.Message):
    position: "TombstoneRespPosition" = betterproto.message_field(
        1, group="position_option"
    )
    no_position: "__client__.Empty" = betterproto.message_field(
        2, group="position_option"
    )


@dataclass(eq=False, repr=False)
class TombstoneRespPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


class StreamsStub(betterproto.ServiceStub):
    async def read(
        self,
        read_req: "ReadReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator["ReadResp"]:
        async for response in self._unary_stream(
            "/event_store.client.streams.Streams/Read",
            read_req,
            ReadResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def append(
        self,
        append_req_iterator: Union[AsyncIterable["AppendReq"], Iterable["AppendReq"]],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "AppendResp":
        return await self._stream_unary(
            "/event_store.client.streams.Streams/Append",
            append_req_iterator,
            AppendReq,
            AppendResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete(
        self,
        delete_req: "DeleteReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "DeleteResp":
        return await self._unary_unary(
            "/event_store.client.streams.Streams/Delete",
            delete_req,
            DeleteResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def tombstone(
        self,
        tombstone_req: "TombstoneReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "TombstoneResp":
        return await self._unary_unary(
            "/event_store.client.streams.Streams/Tombstone",
            tombstone_req,
            TombstoneResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def batch_append(
        self,
        batch_append_req_iterator: Union[
            AsyncIterable["BatchAppendReq"], Iterable["BatchAppendReq"]
        ],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator["BatchAppendResp"]:
        async for response in self._stream_stream(
            "/event_store.client.streams.Streams/BatchAppend",
            batch_append_req_iterator,
            BatchAppendReq,
            BatchAppendResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response


class StreamsBase(ServiceBase):
    async def read(self, read_req: "ReadReq") -> AsyncIterator["ReadResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def append(
        self, append_req_iterator: AsyncIterator["AppendReq"]
    ) -> "AppendResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self, delete_req: "DeleteReq") -> "DeleteResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def tombstone(self, tombstone_req: "TombstoneReq") -> "TombstoneResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def batch_append(
        self, batch_append_req_iterator: AsyncIterator["BatchAppendReq"]
    ) -> AsyncIterator["BatchAppendResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_read(
        self, stream: "grpclib.server.Stream[ReadReq, ReadResp]"
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.read,
            stream,
            request,
        )

    async def __rpc_append(
        self, stream: "grpclib.server.Stream[AppendReq, AppendResp]"
    ) -> None:
        request = stream.__aiter__()
        response = await self.append(request)
        await stream.send_message(response)

    async def __rpc_delete(
        self, stream: "grpclib.server.Stream[DeleteReq, DeleteResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete(request)
        await stream.send_message(response)

    async def __rpc_tombstone(
        self, stream: "grpclib.server.Stream[TombstoneReq, TombstoneResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.tombstone(request)
        await stream.send_message(response)

    async def __rpc_batch_append(
        self, stream: "grpclib.server.Stream[BatchAppendReq, BatchAppendResp]"
    ) -> None:
        request = stream.__aiter__()
        await self._call_rpc_handler_server_stream(
            self.batch_append,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.streams.Streams/Read": grpclib.const.Handler(
                self.__rpc_read,
                grpclib.const.Cardinality.UNARY_STREAM,
                ReadReq,
                ReadResp,
            ),
            "/event_store.client.streams.Streams/Append": grpclib.const.Handler(
                self.__rpc_append,
                grpclib.const.Cardinality.STREAM_UNARY,
                AppendReq,
                AppendResp,
            ),
            "/event_store.client.streams.Streams/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteReq,
                DeleteResp,
            ),
            "/event_store.client.streams.Streams/Tombstone": grpclib.const.Handler(
                self.__rpc_tombstone,
                grpclib.const.Cardinality.UNARY_UNARY,
                TombstoneReq,
                TombstoneResp,
            ),
            "/event_store.client.streams.Streams/BatchAppend": grpclib.const.Handler(
                self.__rpc_batch_append,
                grpclib.const.Cardinality.STREAM_STREAM,
                BatchAppendReq,
                BatchAppendResp,
            ),
        }



from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from eventstoredb.streams.types import (
    AllPosition,
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
    StreamPosition,
    StreamRevision,
)


@dataclass
class SubscribeToStreamOptions:
    from_revision: StreamPosition | StreamRevision = StreamPosition.START
    resolve_links: bool = False


@dataclass
class SubscribeToAllOptions:
    from_position: StreamPosition | AllPosition = StreamPosition.START
    resolve_links: bool = False
    filter: ExcludeSystemEventsFilter | EventTypeFilter | StreamNameFilter | None = None
    max_search_window: int | None = None
    checkpoint_interval: int = 1


@dataclass
class SubscriptionConfirmation:
    id: UUID



import asyncio
from dataclasses import dataclass
from operator import itemgetter
from typing import Awaitable, Callable, List, Optional

import requests

from eventstoredb.events import JsonEvent, ReadEvent
from eventstoredb.persistent_subscriptions.subscribe import PersistentSubscription
from eventstoredb.persistent_subscriptions.subscribe.types import (
    PersistentSubscriptionEvent,
)
from eventstoredb.streams.subscribe import Subscription


def json_test_events(amount: int) -> List[JsonEvent]:
    events = []
    for i in range(amount):
        events.append(JsonEvent(type=f"Test{i+1}"))
    return events


@dataclass
class EventstoreHTTP:
    host: str
    port: int

    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}"

    def read_stream(self, stream_name: str, event_id: Optional[int] = None):
        url = f"{self.url}/streams/{stream_name}"
        if event_id is not None:
            url = f"{url}/{event_id}"
        response = requests.get(
            url=url,
            headers={"Accept": "application/vnd.eventstore.atom+json"},
            params={"embed": "TryHarder"},
        )
        data = response.json()
        if event_id is not None:
            return data
        else:
            events = data.get("entries")
            return sorted(events, key=itemgetter("positionEventNumber"))

    def get_persistent_subscriptions(self, stream_name: Optional[str] = None):
        url = f"{self.url}/subscriptions"
        if stream_name:
            url = f"{url}/{stream_name}"
        response = requests.get(url=url)
        return response.json()

    def get_persistent_subscription_details(self, stream_name: str, group_name: str):
        url = f"{self.url}/subscriptions/{stream_name}/{group_name}/info"
        response = requests.get(url=url)
        return response.json()


class Subscriber:
    def __init__(self, loop):
        self._loop = loop
        self._subscription: Subscription
        self.event_handler: Callable[[ReadEvent], None]

    @property
    def subscription(self):
        return self._subscription

    @subscription.setter
    def subscription(self, value):
        self._subscription = value
        self._task = self._loop.create_task(self._consume())

    async def _consume(self):
        async for event in self._subscription:
            self.event_handler(event)

    async def _stop(self):
        self._task.cancel()
        try:
            await self._task
        except asyncio.CancelledError:
            pass


class PersistentSubscriber:
    def __init__(self, loop):
        self._loop = loop
        self._subscription: PersistentSubscription
        self.event_handler: Callable[[PersistentSubscriptionEvent], Awaitable[None]]

    @property
    def subscription(self):
        return self._subscription

    @subscription.setter
    def subscription(self, value):
        self._subscription = value
        self._task = self._loop.create_task(self._consume())

    async def _consume(self):
        async for event in self._subscription:
            await self.event_handler(event)

    async def _stop(self):
        self._task.cancel()
        try:
            await self._task
        except asyncio.CancelledError:
            pass



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: persistent.proto
# plugin: python-betterproto
import warnings
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from ... import client as __client__

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class ReadReqNackAction(betterproto.Enum):
    Unknown = 0
    Park = 1
    Retry = 2
    Skip = 3
    Stop = 4


class CreateReqConsumerStrategy(betterproto.Enum):
    DispatchToSingle = 0
    RoundRobin = 1
    Pinned = 2


class UpdateReqConsumerStrategy(betterproto.Enum):
    DispatchToSingle = 0
    RoundRobin = 1
    Pinned = 2


@dataclass(eq=False, repr=False)
class ReadReq(betterproto.Message):
    options: "ReadReqOptions" = betterproto.message_field(1, group="content")
    ack: "ReadReqAck" = betterproto.message_field(2, group="content")
    nack: "ReadReqNack" = betterproto.message_field(3, group="content")


@dataclass(eq=False, repr=False)
class ReadReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(
        1, group="stream_option"
    )
    all: "__client__.Empty" = betterproto.message_field(5, group="stream_option")
    group_name: str = betterproto.string_field(2)
    buffer_size: int = betterproto.int32_field(3)
    uuid_option: "ReadReqOptionsUuidOption" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class ReadReqOptionsUuidOption(betterproto.Message):
    structured: "__client__.Empty" = betterproto.message_field(1, group="content")
    string: "__client__.Empty" = betterproto.message_field(2, group="content")


@dataclass(eq=False, repr=False)
class ReadReqAck(betterproto.Message):
    id: bytes = betterproto.bytes_field(1)
    ids: List["__client__.Uuid"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class ReadReqNack(betterproto.Message):
    id: bytes = betterproto.bytes_field(1)
    ids: List["__client__.Uuid"] = betterproto.message_field(2)
    action: "ReadReqNackAction" = betterproto.enum_field(3)
    reason: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ReadResp(betterproto.Message):
    event: "ReadRespReadEvent" = betterproto.message_field(1, group="content")
    subscription_confirmation: "ReadRespSubscriptionConfirmation" = (
        betterproto.message_field(2, group="content")
    )


@dataclass(eq=False, repr=False)
class ReadRespReadEvent(betterproto.Message):
    event: "ReadRespReadEventRecordedEvent" = betterproto.message_field(1)
    link: "ReadRespReadEventRecordedEvent" = betterproto.message_field(2)
    commit_position: int = betterproto.uint64_field(3, group="position")
    no_position: "__client__.Empty" = betterproto.message_field(4, group="position")
    retry_count: int = betterproto.int32_field(5, group="count")
    no_retry_count: "__client__.Empty" = betterproto.message_field(6, group="count")


@dataclass(eq=False, repr=False)
class ReadRespReadEventRecordedEvent(betterproto.Message):
    id: "__client__.Uuid" = betterproto.message_field(1)
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(2)
    stream_revision: int = betterproto.uint64_field(3)
    prepare_position: int = betterproto.uint64_field(4)
    commit_position: int = betterproto.uint64_field(5)
    metadata: Dict[str, str] = betterproto.map_field(
        6, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    custom_metadata: bytes = betterproto.bytes_field(7)
    data: bytes = betterproto.bytes_field(8)


@dataclass(eq=False, repr=False)
class ReadRespSubscriptionConfirmation(betterproto.Message):
    subscription_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CreateReq(betterproto.Message):
    options: "CreateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptions(betterproto.Message):
    stream: "CreateReqStreamOptions" = betterproto.message_field(
        4, group="stream_option"
    )
    all: "CreateReqAllOptions" = betterproto.message_field(5, group="stream_option")
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    group_name: str = betterproto.string_field(2)
    settings: "CreateReqSettings" = betterproto.message_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("stream_identifier"):
            warnings.warn(
                "CreateReqOptions.stream_identifier is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class CreateReqStreamOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    revision: int = betterproto.uint64_field(2, group="revision_option")
    start: "__client__.Empty" = betterproto.message_field(3, group="revision_option")
    end: "__client__.Empty" = betterproto.message_field(4, group="revision_option")


@dataclass(eq=False, repr=False)
class CreateReqAllOptions(betterproto.Message):
    position: "CreateReqPosition" = betterproto.message_field(1, group="all_option")
    start: "__client__.Empty" = betterproto.message_field(2, group="all_option")
    end: "__client__.Empty" = betterproto.message_field(3, group="all_option")
    filter: "CreateReqAllOptionsFilterOptions" = betterproto.message_field(
        4, group="filter_option"
    )
    no_filter: "__client__.Empty" = betterproto.message_field(5, group="filter_option")


@dataclass(eq=False, repr=False)
class CreateReqAllOptionsFilterOptions(betterproto.Message):
    stream_identifier: "CreateReqAllOptionsFilterOptionsExpression" = (
        betterproto.message_field(1, group="filter")
    )
    event_type: "CreateReqAllOptionsFilterOptionsExpression" = (
        betterproto.message_field(2, group="filter")
    )
    max: int = betterproto.uint32_field(3, group="window")
    count: "__client__.Empty" = betterproto.message_field(4, group="window")
    checkpoint_interval_multiplier: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class CreateReqAllOptionsFilterOptionsExpression(betterproto.Message):
    regex: str = betterproto.string_field(1)
    prefix: List[str] = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CreateReqPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CreateReqSettings(betterproto.Message):
    resolve_links: bool = betterproto.bool_field(1)
    revision: int = betterproto.uint64_field(2)
    extra_statistics: bool = betterproto.bool_field(3)
    max_retry_count: int = betterproto.int32_field(5)
    min_checkpoint_count: int = betterproto.int32_field(7)
    max_checkpoint_count: int = betterproto.int32_field(8)
    max_subscriber_count: int = betterproto.int32_field(9)
    live_buffer_size: int = betterproto.int32_field(10)
    read_batch_size: int = betterproto.int32_field(11)
    history_buffer_size: int = betterproto.int32_field(12)
    named_consumer_strategy: "CreateReqConsumerStrategy" = betterproto.enum_field(13)
    message_timeout_ticks: int = betterproto.int64_field(4, group="message_timeout")
    message_timeout_ms: int = betterproto.int32_field(14, group="message_timeout")
    checkpoint_after_ticks: int = betterproto.int64_field(6, group="checkpoint_after")
    checkpoint_after_ms: int = betterproto.int32_field(15, group="checkpoint_after")
    consumer_strategy: str = betterproto.string_field(16)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("revision"):
            warnings.warn(
                "CreateReqSettings.revision is deprecated", DeprecationWarning
            )
        if self.is_set("named_consumer_strategy"):
            warnings.warn(
                "CreateReqSettings.named_consumer_strategy is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class CreateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class UpdateReq(betterproto.Message):
    options: "UpdateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UpdateReqOptions(betterproto.Message):
    stream: "UpdateReqStreamOptions" = betterproto.message_field(
        4, group="stream_option"
    )
    all: "UpdateReqAllOptions" = betterproto.message_field(5, group="stream_option")
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    group_name: str = betterproto.string_field(2)
    settings: "UpdateReqSettings" = betterproto.message_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("stream_identifier"):
            warnings.warn(
                "UpdateReqOptions.stream_identifier is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class UpdateReqStreamOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(1)
    revision: int = betterproto.uint64_field(2, group="revision_option")
    start: "__client__.Empty" = betterproto.message_field(3, group="revision_option")
    end: "__client__.Empty" = betterproto.message_field(4, group="revision_option")


@dataclass(eq=False, repr=False)
class UpdateReqAllOptions(betterproto.Message):
    position: "UpdateReqPosition" = betterproto.message_field(1, group="all_option")
    start: "__client__.Empty" = betterproto.message_field(2, group="all_option")
    end: "__client__.Empty" = betterproto.message_field(3, group="all_option")


@dataclass(eq=False, repr=False)
class UpdateReqPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class UpdateReqSettings(betterproto.Message):
    resolve_links: bool = betterproto.bool_field(1)
    revision: int = betterproto.uint64_field(2)
    extra_statistics: bool = betterproto.bool_field(3)
    max_retry_count: int = betterproto.int32_field(5)
    min_checkpoint_count: int = betterproto.int32_field(7)
    max_checkpoint_count: int = betterproto.int32_field(8)
    max_subscriber_count: int = betterproto.int32_field(9)
    live_buffer_size: int = betterproto.int32_field(10)
    read_batch_size: int = betterproto.int32_field(11)
    history_buffer_size: int = betterproto.int32_field(12)
    named_consumer_strategy: "UpdateReqConsumerStrategy" = betterproto.enum_field(13)
    message_timeout_ticks: int = betterproto.int64_field(4, group="message_timeout")
    message_timeout_ms: int = betterproto.int32_field(14, group="message_timeout")
    checkpoint_after_ticks: int = betterproto.int64_field(6, group="checkpoint_after")
    checkpoint_after_ms: int = betterproto.int32_field(15, group="checkpoint_after")

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("revision"):
            warnings.warn(
                "UpdateReqSettings.revision is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class UpdateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DeleteReq(betterproto.Message):
    options: "DeleteReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(
        1, group="stream_option"
    )
    all: "__client__.Empty" = betterproto.message_field(3, group="stream_option")
    group_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class DeleteResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetInfoReq(betterproto.Message):
    options: "GetInfoReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetInfoReqOptions(betterproto.Message):
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(
        1, group="stream_option"
    )
    all: "__client__.Empty" = betterproto.message_field(2, group="stream_option")
    group_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class GetInfoResp(betterproto.Message):
    subscription_info: "SubscriptionInfo" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class SubscriptionInfo(betterproto.Message):
    event_source: str = betterproto.string_field(1)
    group_name: str = betterproto.string_field(2)
    status: str = betterproto.string_field(3)
    connections: List["SubscriptionInfoConnectionInfo"] = betterproto.message_field(4)
    average_per_second: int = betterproto.int32_field(5)
    total_items: int = betterproto.int64_field(6)
    count_since_last_measurement: int = betterproto.int64_field(7)
    last_checkpointed_event_position: str = betterproto.string_field(8)
    last_known_event_position: str = betterproto.string_field(9)
    resolve_link_tos: bool = betterproto.bool_field(10)
    start_from: str = betterproto.string_field(11)
    message_timeout_milliseconds: int = betterproto.int32_field(12)
    extra_statistics: bool = betterproto.bool_field(13)
    max_retry_count: int = betterproto.int32_field(14)
    live_buffer_size: int = betterproto.int32_field(15)
    buffer_size: int = betterproto.int32_field(16)
    read_batch_size: int = betterproto.int32_field(17)
    check_point_after_milliseconds: int = betterproto.int32_field(18)
    min_check_point_count: int = betterproto.int32_field(19)
    max_check_point_count: int = betterproto.int32_field(20)
    read_buffer_count: int = betterproto.int32_field(21)
    live_buffer_count: int = betterproto.int64_field(22)
    retry_buffer_count: int = betterproto.int32_field(23)
    total_in_flight_messages: int = betterproto.int32_field(24)
    outstanding_messages_count: int = betterproto.int32_field(25)
    named_consumer_strategy: str = betterproto.string_field(26)
    max_subscriber_count: int = betterproto.int32_field(27)
    parked_message_count: int = betterproto.int64_field(28)


@dataclass(eq=False, repr=False)
class SubscriptionInfoConnectionInfo(betterproto.Message):
    from_: str = betterproto.string_field(1)
    username: str = betterproto.string_field(2)
    average_items_per_second: int = betterproto.int32_field(3)
    total_items: int = betterproto.int64_field(4)
    count_since_last_measurement: int = betterproto.int64_field(5)
    observed_measurements: List[
        "SubscriptionInfoMeasurement"
    ] = betterproto.message_field(6)
    available_slots: int = betterproto.int32_field(7)
    in_flight_messages: int = betterproto.int32_field(8)
    connection_name: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class SubscriptionInfoMeasurement(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: int = betterproto.int64_field(2)


@dataclass(eq=False, repr=False)
class ReplayParkedReq(betterproto.Message):
    options: "ReplayParkedReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ReplayParkedReqOptions(betterproto.Message):
    group_name: str = betterproto.string_field(1)
    stream_identifier: "__client__.StreamIdentifier" = betterproto.message_field(
        2, group="stream_option"
    )
    all: "__client__.Empty" = betterproto.message_field(3, group="stream_option")
    stop_at: int = betterproto.int64_field(4, group="stop_at_option")
    no_limit: "__client__.Empty" = betterproto.message_field(5, group="stop_at_option")


@dataclass(eq=False, repr=False)
class ReplayParkedResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ListReq(betterproto.Message):
    options: "ListReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ListReqOptions(betterproto.Message):
    list_all_subscriptions: "__client__.Empty" = betterproto.message_field(
        1, group="list_option"
    )
    list_for_stream: "ListReqStreamOption" = betterproto.message_field(
        2, group="list_option"
    )


@dataclass(eq=False, repr=False)
class ListReqStreamOption(betterproto.Message):
    stream: "__client__.StreamIdentifier" = betterproto.message_field(
        1, group="stream_option"
    )
    all: "__client__.Empty" = betterproto.message_field(2, group="stream_option")


@dataclass(eq=False, repr=False)
class ListResp(betterproto.Message):
    subscriptions: List["SubscriptionInfo"] = betterproto.message_field(1)


class PersistentSubscriptionsStub(betterproto.ServiceStub):
    async def create(
        self,
        create_req: "CreateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "CreateResp":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Create",
            create_req,
            CreateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def update(
        self,
        update_req: "UpdateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "UpdateResp":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Update",
            update_req,
            UpdateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete(
        self,
        delete_req: "DeleteReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "DeleteResp":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Delete",
            delete_req,
            DeleteResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def read(
        self,
        read_req_iterator: Union[AsyncIterable["ReadReq"], Iterable["ReadReq"]],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator["ReadResp"]:
        async for response in self._stream_stream(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Read",
            read_req_iterator,
            ReadReq,
            ReadResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def get_info(
        self,
        get_info_req: "GetInfoReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "GetInfoResp":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/GetInfo",
            get_info_req,
            GetInfoResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def replay_parked(
        self,
        replay_parked_req: "ReplayParkedReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ReplayParkedResp":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/ReplayParked",
            replay_parked_req,
            ReplayParkedResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def list(
        self,
        list_req: "ListReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ListResp":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/List",
            list_req,
            ListResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def restart_subsystem(
        self,
        client_empty: "__client__.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "__client__.Empty":
        return await self._unary_unary(
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/RestartSubsystem",
            client_empty,
            __client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class PersistentSubscriptionsBase(ServiceBase):
    async def create(self, create_req: "CreateReq") -> "CreateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def update(self, update_req: "UpdateReq") -> "UpdateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self, delete_req: "DeleteReq") -> "DeleteResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def read(
        self, read_req_iterator: AsyncIterator["ReadReq"]
    ) -> AsyncIterator["ReadResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_info(self, get_info_req: "GetInfoReq") -> "GetInfoResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def replay_parked(
        self, replay_parked_req: "ReplayParkedReq"
    ) -> "ReplayParkedResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def list(self, list_req: "ListReq") -> "ListResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def restart_subsystem(
        self, client_empty: "__client__.Empty"
    ) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_create(
        self, stream: "grpclib.server.Stream[CreateReq, CreateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.create(request)
        await stream.send_message(response)

    async def __rpc_update(
        self, stream: "grpclib.server.Stream[UpdateReq, UpdateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.update(request)
        await stream.send_message(response)

    async def __rpc_delete(
        self, stream: "grpclib.server.Stream[DeleteReq, DeleteResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete(request)
        await stream.send_message(response)

    async def __rpc_read(
        self, stream: "grpclib.server.Stream[ReadReq, ReadResp]"
    ) -> None:
        request = stream.__aiter__()
        await self._call_rpc_handler_server_stream(
            self.read,
            stream,
            request,
        )

    async def __rpc_get_info(
        self, stream: "grpclib.server.Stream[GetInfoReq, GetInfoResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_info(request)
        await stream.send_message(response)

    async def __rpc_replay_parked(
        self, stream: "grpclib.server.Stream[ReplayParkedReq, ReplayParkedResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.replay_parked(request)
        await stream.send_message(response)

    async def __rpc_list(
        self, stream: "grpclib.server.Stream[ListReq, ListResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.list(request)
        await stream.send_message(response)

    async def __rpc_restart_subsystem(
        self, stream: "grpclib.server.Stream[__client__.Empty, __client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.restart_subsystem(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Create": grpclib.const.Handler(
                self.__rpc_create,
                grpclib.const.Cardinality.UNARY_UNARY,
                CreateReq,
                CreateResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                UpdateReq,
                UpdateResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteReq,
                DeleteResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/Read": grpclib.const.Handler(
                self.__rpc_read,
                grpclib.const.Cardinality.STREAM_STREAM,
                ReadReq,
                ReadResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/GetInfo": grpclib.const.Handler(
                self.__rpc_get_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetInfoReq,
                GetInfoResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/ReplayParked": grpclib.const.Handler(
                self.__rpc_replay_parked,
                grpclib.const.Cardinality.UNARY_UNARY,
                ReplayParkedReq,
                ReplayParkedResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/List": grpclib.const.Handler(
                self.__rpc_list,
                grpclib.const.Cardinality.UNARY_UNARY,
                ListReq,
                ListResp,
            ),
            "/event_store.client.persistent_subscriptions.PersistentSubscriptions/RestartSubsystem": grpclib.const.Handler(
                self.__rpc_restart_subsystem,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
        }



import asyncio
from typing import AsyncGenerator

import pytest
from pytest_mock import MockerFixture

from eventstoredb import Client
from eventstoredb.events import JsonEvent
from eventstoredb.streams.subscribe import SubscribeToStreamOptions
from eventstoredb.streams.types import StreamPosition

from ..utils import Subscriber, json_test_events


@pytest.fixture
async def subscriber(
    event_loop: asyncio.AbstractEventLoop,
    mocker: MockerFixture,
) -> AsyncGenerator[Subscriber, None]:
    s = Subscriber(event_loop)
    s.event_handler = mocker.stub()
    yield s
    await s._stop()


async def test_subscribe_to_stream_json(
    eventstoredb_client: Client,
    stream_name: str,
    subscriber: Subscriber,
) -> None:
    subscription = eventstoredb_client.subscribe_to_stream(stream_name=stream_name)

    subscriber.subscription = subscription

    assert subscriber.event_handler.call_count == 0

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 1


async def test_subscribe_to_stream_from_start(
    eventstoredb_client: Client,
    stream_name: str,
    subscriber: Subscriber,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    subscription = eventstoredb_client.subscribe_to_stream(
        stream_name=stream_name,
        options=SubscribeToStreamOptions(from_revision=StreamPosition.START),
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 4

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 5


async def test_subscribe_to_stream_from_end(
    eventstoredb_client: Client,
    stream_name: str,
    subscriber: Subscriber,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    subscription = eventstoredb_client.subscribe_to_stream(
        stream_name=stream_name,
        options=SubscribeToStreamOptions(from_revision=StreamPosition.END),
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 0

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 1


async def test_subscribe_to_stream_from_revision(
    eventstoredb_client: Client,
    stream_name: str,
    subscriber: Subscriber,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    subscription = eventstoredb_client.subscribe_to_stream(
        stream_name=stream_name,
        options=SubscribeToStreamOptions(from_revision=2),
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 1

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await asyncio.sleep(0.1)
    assert subscriber.event_handler.call_count == 2



from __future__ import annotations

from typing import Type
from uuid import UUID

import betterproto

from eventstoredb.events import (
    BinaryRecordedEvent,
    ContentType,
    JsonRecordedEvent,
    Position,
    ReadEvent,
)
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.streams import (
    ReadReq,
    ReadReqOptions,
    ReadReqOptionsAllOptions,
    ReadReqOptionsFilterOptions,
    ReadReqOptionsFilterOptionsExpression,
    ReadReqOptionsPosition,
    ReadReqOptionsReadDirection,
    ReadReqOptionsStreamOptions,
    ReadReqOptionsUuidOption,
    ReadResp,
    ReadRespReadEvent,
    ReadRespReadEventRecordedEvent,
)
from eventstoredb.streams.read.exceptions import StreamNotFoundError
from eventstoredb.streams.read.types import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    ReadAllOptions,
    ReadDirection,
    ReadStreamOptions,
    StreamNameFilter,
)
from eventstoredb.streams.types import AllPosition, StreamPosition, StreamRevision


def create_read_request_options_common(
    options: ReadAllOptions | ReadStreamOptions,
) -> ReadReqOptions:
    request_options = ReadReqOptions()

    request_options.resolve_links = options.resolve_links
    request_options.count = options.max_count
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())

    if options.direction == ReadDirection.FORWARDS:
        request_options.read_direction = ReadReqOptionsReadDirection.Forwards
    elif options.direction == ReadDirection.BACKWARDS:
        request_options.read_direction = ReadReqOptionsReadDirection.Backwards

    return request_options


def create_read_request(
    stream_name: str,
    options: ReadStreamOptions | None = None,
) -> ReadReq:
    if options is None:
        options = ReadStreamOptions()

    request_options = create_read_request_options_common(options)

    request_options.no_filter = Empty()
    request_options.stream = ReadReqOptionsStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    return ReadReq(options=request_options)


def create_read_all_request(options: ReadAllOptions | None = None) -> ReadReq:
    if options is None:
        options = ReadAllOptions()

    request_options = create_read_request_options_common(options)

    request_options.all = ReadReqOptionsAllOptions()

    if isinstance(options.from_position, AllPosition):
        request_options.all.position = ReadReqOptionsPosition()
        request_options.all.position.commit_position = options.from_position.commit
        request_options.all.position.prepare_position = options.from_position.prepare
    elif options.from_position == StreamPosition.START:
        request_options.all.start = Empty()
    elif options.from_position == StreamPosition.END:
        request_options.all.end = Empty()

    if not options.filter:
        request_options.no_filter = Empty()
    else:
        request_options.filter = ReadReqOptionsFilterOptions()
        filter_expression = ReadReqOptionsFilterOptionsExpression()

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            filter_expression.regex = "^[^\\$].*"
        else:
            if options.filter.regex:
                filter_expression.regex = options.filter.regex
            if options.filter.prefix:
                filter_expression.prefix = options.filter.prefix

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, EventTypeFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, StreamNameFilter):
            request_options.filter.stream_identifier = filter_expression

        request_options.filter.max = 0
        request_options.filter.count = Empty()

    return ReadReq(options=request_options)


def convert_read_response(message: ReadResp) -> ReadEvent:
    content_type, _ = betterproto.which_one_of(message, "content")
    if content_type == "stream_not_found":
        raise StreamNotFoundError(
            stream_name=message.stream_not_found.stream_identifier.stream_name.decode()
        )
    elif content_type == "event":
        return convert_read_response_read_event(message.event)
    else:
        # FIXME maybe we should raise something like "UnexpectedRuntimeError" here?
        raise Exception("i shouldnt be here")


def convert_read_response_read_event(message: ReadRespReadEvent) -> ReadEvent:
    event = ReadEvent()
    if message.event:
        event.event = convert_read_response_recorded_event(message.event)
    if message.link:
        event.link = convert_read_response_recorded_event(message.link)
    # TODO should this use which_one_of?
    if message.commit_position:
        event.commit_position = message.commit_position
    return event


def convert_read_response_recorded_event(
    message: ReadRespReadEventRecordedEvent,
) -> JsonRecordedEvent | BinaryRecordedEvent:
    stream_name = message.stream_identifier.stream_name.decode()
    id = UUID(message.id.string)
    content_type = ContentType(message.metadata["content-type"])
    position = Position(
        commit=message.commit_position,
        prepare=message.prepare_position,
    )
    event_class: Type[JsonRecordedEvent] | Type[BinaryRecordedEvent]
    if content_type == ContentType.JSON:
        event_class = JsonRecordedEvent
    else:
        event_class = BinaryRecordedEvent
    return event_class(
        stream_name=stream_name,
        id=id,
        revision=message.stream_revision,
        type=message.metadata["type"],
        content_type=content_type,
        created=int(message.metadata["created"]),
        position=position,
        data=message.data if message.data else None,
        metadata=message.custom_metadata if message.custom_metadata else None,
    )



import pytest

from eventstoredb import ClientOptions


@pytest.mark.parametrize(
    "connection_string,client_options",
    [
        (
            "esdb://host",
            ClientOptions(
                host="host",
                port=2113,
                tls=True,
                dns_discovery=False,
            ),
        ),
        (
            "esdb://host:1234",
            ClientOptions(
                host="host",
                port=1234,
                tls=True,
                dns_discovery=False,
            ),
        ),
        (
            "esdb+discovery://host",
            ClientOptions(
                host="host",
                tls=True,
                dns_discovery=True,
            ),
        ),
        (
            "esdb://user:pass@host",
            ClientOptions(
                host="host",
                username="user",
                password="pass",
            ),
        ),
    ],
)
def test_from_url(connection_string: str, client_options: ClientOptions) -> None:
    assert ClientOptions.from_connection_string(connection_string) == client_options



from __future__ import annotations

from eventstoredb.generated.event_store.client import StreamIdentifier
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    DeleteReq,
    DeleteReqOptions,
)
from eventstoredb.persistent_subscriptions.delete.types import (
    DeletePersistentSubscriptionOptions,
)


def create_delete_request(
    stream_name: str,
    group_name: str,
    options: DeletePersistentSubscriptionOptions | None = None,
) -> DeleteReq:
    if options is None:
        options = DeletePersistentSubscriptionOptions()

    request_options = DeleteReqOptions()
    request_options.stream_identifier = StreamIdentifier(stream_name.encode())
    request_options.group_name = group_name

    return DeleteReq(options=request_options)



import uuid

import pytest
import requests
from requests.exceptions import ConnectionError

from eventstoredb import Client, ClientOptions

from .utils import EventstoreHTTP


def pytest_addoption(parser):
    parser.addoption(
        "--skip-docker-services",
        action="store_true",
        help="Skip tests that require docker-services",
    )


@pytest.fixture(autouse=True, scope="session")
def skip_docker_services(request: pytest.FixtureRequest):
    if not request.config.option.skip_docker_services:
        return
    if "docker_services" in request.fixturenames:
        pytest.skip("skipped when not running docker-services")


def eventstoredb_is_ready(url):
    try:
        response = requests.get(f"{url}/stats")
        if response.status_code == 200:
            return True
    except ConnectionError:
        return False


@pytest.fixture(scope="session")
def eventstoredb_service(docker_ip, docker_services):
    port = docker_services.port_for("eventstore.db", 2113)
    url = "http://{}:{}".format(docker_ip, port)
    docker_services.wait_until_responsive(
        timeout=30.0, pause=0.1, check=lambda: eventstoredb_is_ready(url)
    )
    return url


@pytest.fixture
def eventstoredb_host(eventstoredb_service, docker_ip):
    return docker_ip


@pytest.fixture
def eventstoredb_port(eventstoredb_service, docker_services):
    return docker_services.port_for("eventstore.db", 2113)


@pytest.fixture
def eventstoredb_client(eventstoredb_host, eventstoredb_port):
    options = ClientOptions(host=eventstoredb_host, port=eventstoredb_port)
    client = Client(options=options)
    yield client


@pytest.fixture
def eventstoredb_httpclient(eventstoredb_host, eventstoredb_port):
    client = EventstoreHTTP(host=eventstoredb_host, port=eventstoredb_port)
    yield client


@pytest.fixture
def stream_name():
    return f"test-stream-{uuid.uuid4()}"


@pytest.fixture
def group_name():
    return f"test-group-{uuid.uuid4()}"



class PersistentSubscriptionError(Exception):
    pass


class PersistentSubscriptionFailedError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionDoesNotExistError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionExistsError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionDroppedError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionMaxSubscribersReachedError(PersistentSubscriptionError):
    pass



from eventstoredb.persistent_subscriptions.subscribe.grpc import create_read_request
from eventstoredb.persistent_subscriptions.subscribe.subscription import (
    PersistentSubscription,
)
from eventstoredb.persistent_subscriptions.subscribe.types import (
    SubscribeToPersistentSubscriptionOptions,
)



from dataclasses import dataclass

from eventstoredb.persistent_subscriptions.common.types import (
    CreateUpdatePersistentSubscriptionOptions,
)


@dataclass
class UpdatePersistentSubscriptionOptions(CreateUpdatePersistentSubscriptionOptions):
    pass



from __future__ import annotations

from typing import AsyncGenerator, AsyncIterator, Iterable

from grpclib.client import Channel
from grpclib.exceptions import GRPCError

from eventstoredb.client.options import ClientOptions
from eventstoredb.events import EventData, ReadEvent
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    PersistentSubscriptionsStub,
)
from eventstoredb.generated.event_store.client.streams import StreamsStub
from eventstoredb.persistent_subscriptions.common import convert_grpc_error_to_exception
from eventstoredb.persistent_subscriptions.create import (
    CreatePersistentSubscriptionOptions,
    create_create_request,
)
from eventstoredb.persistent_subscriptions.delete import (
    DeletePersistentSubscriptionOptions,
    create_delete_request,
)
from eventstoredb.persistent_subscriptions.subscribe import (
    PersistentSubscription,
    SubscribeToPersistentSubscriptionOptions,
)
from eventstoredb.persistent_subscriptions.update import (
    UpdatePersistentSubscriptionOptions,
    create_update_request,
)
from eventstoredb.streams.append import (
    AppendReq,
    AppendResult,
    AppendToStreamOptions,
    convert_append_response,
    create_append_header,
    create_append_request,
)
from eventstoredb.streams.read import (
    ReadAllOptions,
    ReadStreamOptions,
    convert_read_response,
    create_read_all_request,
    create_read_request,
)
from eventstoredb.streams.subscribe import (
    SubscribeToAllOptions,
    SubscribeToStreamOptions,
    Subscription,
    create_subscribe_to_all_request,
    create_subscribe_to_stream_request,
)


class Client:
    def __init__(self, options: ClientOptions | str) -> None:
        if isinstance(options, str):
            self.options = ClientOptions.from_connection_string(options)
        else:
            self.options = options
        self._channel: Channel | None = None

    @property
    def channel(self) -> Channel:
        if self._channel is None:
            self._channel = Channel(host=self.options.host, port=self.options.port)
        return self._channel

    async def append_to_stream(
        self,
        stream_name: str,
        events: EventData | Iterable[EventData],
        options: AppendToStreamOptions | None = None,
    ) -> AppendResult:
        async def request_iterator() -> AsyncGenerator[AppendReq, None]:
            yield create_append_header(
                stream_name=stream_name,
                options=options,
            )
            if isinstance(events, EventData):
                yield create_append_request(events)
            else:
                for event in events:
                    yield create_append_request(event)

        client = StreamsStub(channel=self.channel)
        response = await client.append(request_iterator())
        return convert_append_response(stream_name, response)

    async def read_stream(
        self,
        stream_name: str,
        options: ReadStreamOptions | None = None,
    ) -> AsyncIterator[ReadEvent]:
        client = StreamsStub(channel=self.channel)
        request = create_read_request(
            stream_name=stream_name,
            options=options,
        )
        # TODO raise exception StreamNotFoundError
        async for response in client.read(read_req=request):
            yield convert_read_response(response)

    async def read_all(
        self,
        options: ReadAllOptions | None = None,
    ) -> AsyncIterator[ReadEvent]:
        client = StreamsStub(channel=self.channel)
        request = create_read_all_request(
            options=options,
        )
        # TODO raise exception StreamNotFoundError
        async for response in client.read(read_req=request):
            yield convert_read_response(response)

    def subscribe_to_stream(
        self,
        stream_name: str,
        options: SubscribeToStreamOptions | None = None,
    ) -> Subscription:
        client = StreamsStub(channel=self.channel)
        request = create_subscribe_to_stream_request(
            stream_name=stream_name,
            options=options,
        )
        it = client.read(read_req=request)
        return Subscription(it)

    def subscribe_to_all(
        self,
        options: SubscribeToAllOptions | None = None,
    ) -> Subscription:
        client = StreamsStub(channel=self.channel)
        request = create_subscribe_to_all_request(options=options)
        it = client.read(read_req=request)
        return Subscription(it)

    async def create_persistent_subscription(
        self,
        stream_name: str,
        group_name: str,
        options: CreatePersistentSubscriptionOptions | None = None,
    ) -> None:
        client = PersistentSubscriptionsStub(channel=self.channel)
        request = create_create_request(
            stream_name=stream_name,
            group_name=group_name,
            options=options,
        )
        try:
            await client.create(create_req=request)
        except GRPCError as e:
            raise convert_grpc_error_to_exception(e)

    async def update_persistent_subscription(
        self,
        stream_name: str,
        group_name: str,
        options: UpdatePersistentSubscriptionOptions | None = None,
    ) -> None:
        client = PersistentSubscriptionsStub(channel=self.channel)
        request = create_update_request(
            stream_name=stream_name,
            group_name=group_name,
            options=options,
        )
        try:
            await client.update(update_req=request)
        except GRPCError as e:
            raise convert_grpc_error_to_exception(e)

    async def delete_persistent_subscription(
        self,
        stream_name: str,
        group_name: str,
        options: DeletePersistentSubscriptionOptions | None = None,
    ) -> None:
        client = PersistentSubscriptionsStub(channel=self.channel)
        request = create_delete_request(
            stream_name=stream_name,
            group_name=group_name,
            options=options,
        )
        try:
            await client.delete(delete_req=request)
        except GRPCError as e:
            raise convert_grpc_error_to_exception(e)

    def subscribe_to_persistent_subscription(
        self,
        stream_name: str,
        group_name: str,
        options: SubscribeToPersistentSubscriptionOptions | None = None,
    ) -> PersistentSubscription:
        client = PersistentSubscriptionsStub(channel=self.channel)
        return PersistentSubscription(
            client=client,
            stream_name=stream_name,
            group_name=group_name,
            options=options,
        )



from __future__ import annotations

from uuid import UUID

from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.streams import (
    ReadReq,
    ReadReqOptions,
    ReadReqOptionsAllOptions,
    ReadReqOptionsFilterOptions,
    ReadReqOptionsFilterOptionsExpression,
    ReadReqOptionsPosition,
    ReadReqOptionsStreamOptions,
    ReadReqOptionsSubscriptionOptions,
    ReadReqOptionsUuidOption,
    ReadRespSubscriptionConfirmation,
)
from eventstoredb.streams.subscribe.types import (
    SubscribeToAllOptions,
    SubscribeToStreamOptions,
    SubscriptionConfirmation,
)
from eventstoredb.streams.types import (
    AllPosition,
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
    StreamPosition,
    StreamRevision,
)


def create_read_request_options_common(
    options: SubscribeToAllOptions | SubscribeToStreamOptions,
) -> ReadReqOptions:
    request_options = ReadReqOptions()

    request_options.resolve_links = options.resolve_links
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())
    request_options.subscription = ReadReqOptionsSubscriptionOptions()

    return request_options


def create_subscribe_to_stream_request(
    stream_name: str,
    options: SubscribeToStreamOptions | None = None,
) -> ReadReq:
    if options is None:
        options = SubscribeToStreamOptions()

    request_options = create_read_request_options_common(options)

    request_options.no_filter = Empty()

    request_options.stream = ReadReqOptionsStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    return ReadReq(options=request_options)


def create_subscribe_to_all_request(
    options: SubscribeToAllOptions | None = None,
) -> ReadReq:
    if options is None:
        options = SubscribeToAllOptions()

    request_options = create_read_request_options_common(options)

    request_options.all = ReadReqOptionsAllOptions()

    if isinstance(options.from_position, AllPosition):
        request_options.all.position = ReadReqOptionsPosition()
        request_options.all.position.commit_position = options.from_position.commit
        request_options.all.position.prepare_position = options.from_position.prepare
    elif options.from_position == StreamPosition.START:
        request_options.all.start = Empty()
    elif options.from_position == StreamPosition.END:
        request_options.all.end = Empty()

    if not options.filter:
        request_options.no_filter = Empty()
    else:
        request_options.filter = ReadReqOptionsFilterOptions()
        filter_expression = ReadReqOptionsFilterOptionsExpression()

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            filter_expression.regex = "^[^\\$].*"
        else:
            if options.filter.regex:
                filter_expression.regex = options.filter.regex
            if options.filter.prefix:
                filter_expression.prefix = options.filter.prefix

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, EventTypeFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, StreamNameFilter):
            request_options.filter.stream_identifier = filter_expression

        if options.max_search_window is not None:
            request_options.filter.max = options.max_search_window
        else:
            request_options.filter.count = Empty()

        request_options.filter.checkpoint_interval_multiplier = (
            options.checkpoint_interval
        )

    return ReadReq(options=request_options)


def convert_read_response_confirmation(
    message: ReadRespSubscriptionConfirmation,
) -> SubscriptionConfirmation:
    return SubscriptionConfirmation(id=UUID(message.subscription_id))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: cluster.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import client as _client__

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class MemberInfoVNodeState(betterproto.Enum):
    Initializing = 0
    DiscoverLeader = 1
    Unknown = 2
    PreReplica = 3
    CatchingUp = 4
    Clone = 5
    Follower = 6
    PreLeader = 7
    Leader = 8
    Manager = 9
    ShuttingDown = 10
    Shutdown = 11
    ReadOnlyLeaderless = 12
    PreReadOnlyReplica = 13
    ReadOnlyReplica = 14
    ResigningLeader = 15


@dataclass(eq=False, repr=False)
class GossipRequest(betterproto.Message):
    info: "ClusterInfo" = betterproto.message_field(1)
    server: "EndPoint" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class ViewChangeRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    attempted_view: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class ViewChangeProofRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    installed_view: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class PrepareRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    view: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class PrepareOkRequest(betterproto.Message):
    view: int = betterproto.int32_field(1)
    server_id: "_client__.Uuid" = betterproto.message_field(2)
    server_http: "EndPoint" = betterproto.message_field(3)
    epoch_number: int = betterproto.int32_field(4)
    epoch_position: int = betterproto.int64_field(5)
    epoch_id: "_client__.Uuid" = betterproto.message_field(6)
    epoch_leader_instance_id: "_client__.Uuid" = betterproto.message_field(7)
    last_commit_position: int = betterproto.int64_field(8)
    writer_checkpoint: int = betterproto.int64_field(9)
    chaser_checkpoint: int = betterproto.int64_field(10)
    node_priority: int = betterproto.int32_field(11)
    cluster_info: "ClusterInfo" = betterproto.message_field(12)


@dataclass(eq=False, repr=False)
class ProposalRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    leader_id: "_client__.Uuid" = betterproto.message_field(3)
    leader_http: "EndPoint" = betterproto.message_field(4)
    view: int = betterproto.int32_field(5)
    epoch_number: int = betterproto.int32_field(6)
    epoch_position: int = betterproto.int64_field(7)
    epoch_id: "_client__.Uuid" = betterproto.message_field(8)
    epoch_leader_instance_id: "_client__.Uuid" = betterproto.message_field(9)
    last_commit_position: int = betterproto.int64_field(10)
    writer_checkpoint: int = betterproto.int64_field(11)
    chaser_checkpoint: int = betterproto.int64_field(12)
    node_priority: int = betterproto.int32_field(13)


@dataclass(eq=False, repr=False)
class AcceptRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    leader_id: "_client__.Uuid" = betterproto.message_field(3)
    leader_http: "EndPoint" = betterproto.message_field(4)
    view: int = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class LeaderIsResigningRequest(betterproto.Message):
    leader_id: "_client__.Uuid" = betterproto.message_field(1)
    leader_http: "EndPoint" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class LeaderIsResigningOkRequest(betterproto.Message):
    leader_id: "_client__.Uuid" = betterproto.message_field(1)
    leader_http: "EndPoint" = betterproto.message_field(2)
    server_id: "_client__.Uuid" = betterproto.message_field(3)
    server_http: "EndPoint" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class ClusterInfo(betterproto.Message):
    members: List["MemberInfo"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class EndPoint(betterproto.Message):
    address: str = betterproto.string_field(1)
    port: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class MemberInfo(betterproto.Message):
    instance_id: "_client__.Uuid" = betterproto.message_field(1)
    time_stamp: int = betterproto.int64_field(2)
    state: "MemberInfoVNodeState" = betterproto.enum_field(3)
    is_alive: bool = betterproto.bool_field(4)
    http_end_point: "EndPoint" = betterproto.message_field(5)
    internal_tcp: "EndPoint" = betterproto.message_field(6)
    external_tcp: "EndPoint" = betterproto.message_field(7)
    internal_tcp_uses_tls: bool = betterproto.bool_field(8)
    external_tcp_uses_tls: bool = betterproto.bool_field(9)
    last_commit_position: int = betterproto.int64_field(10)
    writer_checkpoint: int = betterproto.int64_field(11)
    chaser_checkpoint: int = betterproto.int64_field(12)
    epoch_position: int = betterproto.int64_field(13)
    epoch_number: int = betterproto.int32_field(14)
    epoch_id: "_client__.Uuid" = betterproto.message_field(15)
    node_priority: int = betterproto.int32_field(16)
    is_read_only_replica: bool = betterproto.bool_field(17)
    advertise_host_to_client_as: str = betterproto.string_field(18)
    advertise_http_port_to_client_as: int = betterproto.uint32_field(19)
    advertise_tcp_port_to_client_as: int = betterproto.uint32_field(20)


@dataclass(eq=False, repr=False)
class ReplicaLogWrite(betterproto.Message):
    log_position: int = betterproto.int64_field(1)
    replica_id: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class ReplicatedTo(betterproto.Message):
    log_position: int = betterproto.int64_field(1)


@dataclass(eq=False, repr=False)
class Epoch(betterproto.Message):
    epoch_position: int = betterproto.int64_field(1)
    epoch_number: int = betterproto.int32_field(2)
    epoch_id: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class SubscribeReplica(betterproto.Message):
    log_position: int = betterproto.int64_field(1)
    chunk_id: bytes = betterproto.bytes_field(2)
    last_epochs: List["Epoch"] = betterproto.message_field(3)
    ip: bytes = betterproto.bytes_field(4)
    port: int = betterproto.int32_field(5)
    leader_id: bytes = betterproto.bytes_field(6)
    subscription_id: bytes = betterproto.bytes_field(7)
    is_promotable: bool = betterproto.bool_field(8)
    version: int = betterproto.int32_field(9)


@dataclass(eq=False, repr=False)
class ReplicaSubscriptionRetry(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class ReplicaSubscribed(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)
    subscription_position: int = betterproto.int64_field(3)


@dataclass(eq=False, repr=False)
class ReplicaLogPositionAck(betterproto.Message):
    subscription_id: bytes = betterproto.bytes_field(1)
    replication_log_position: int = betterproto.int64_field(2)
    writer_log_position: int = betterproto.int64_field(3)


@dataclass(eq=False, repr=False)
class CreateChunk(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)
    chunk_header_bytes: bytes = betterproto.bytes_field(3)
    file_size: int = betterproto.int32_field(4)
    is_completed_chunk: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class RawChunkBulk(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)
    chunk_start_number: int = betterproto.int32_field(3)
    chunk_end_number: int = betterproto.int32_field(4)
    raw_position: int = betterproto.int32_field(5)
    raw_bytes: bytes = betterproto.bytes_field(6)
    complete_chunk: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class DataChunkBulk(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)
    chunk_start_number: int = betterproto.int32_field(3)
    chunk_end_number: int = betterproto.int32_field(4)
    subscription_position: int = betterproto.int64_field(5)
    data_bytes: bytes = betterproto.bytes_field(6)
    complete_chunk: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class FollowerAssignment(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CloneAssignment(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class DropSubscription(betterproto.Message):
    leader_id: bytes = betterproto.bytes_field(1)
    subscription_id: bytes = betterproto.bytes_field(2)


class GossipStub(betterproto.ServiceStub):
    async def update(
        self,
        gossip_request: "GossipRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ClusterInfo":
        return await self._unary_unary(
            "/event_store.cluster.Gossip/Update",
            gossip_request,
            ClusterInfo,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def read(
        self,
        client_empty: "_client__.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ClusterInfo":
        return await self._unary_unary(
            "/event_store.cluster.Gossip/Read",
            client_empty,
            ClusterInfo,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class ElectionsStub(betterproto.ServiceStub):
    async def view_change(
        self,
        view_change_request: "ViewChangeRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/ViewChange",
            view_change_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def view_change_proof(
        self,
        view_change_proof_request: "ViewChangeProofRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/ViewChangeProof",
            view_change_proof_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def prepare(
        self,
        prepare_request: "PrepareRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/Prepare",
            prepare_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def prepare_ok(
        self,
        prepare_ok_request: "PrepareOkRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/PrepareOk",
            prepare_ok_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def proposal(
        self,
        proposal_request: "ProposalRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/Proposal",
            proposal_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def accept(
        self,
        accept_request: "AcceptRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/Accept",
            accept_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def leader_is_resigning(
        self,
        leader_is_resigning_request: "LeaderIsResigningRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/LeaderIsResigning",
            leader_is_resigning_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def leader_is_resigning_ok(
        self,
        leader_is_resigning_ok_request: "LeaderIsResigningOkRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "_client__.Empty":
        return await self._unary_unary(
            "/event_store.cluster.Elections/LeaderIsResigningOk",
            leader_is_resigning_ok_request,
            _client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class GossipBase(ServiceBase):
    async def update(self, gossip_request: "GossipRequest") -> "ClusterInfo":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def read(self, client_empty: "_client__.Empty") -> "ClusterInfo":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_update(
        self, stream: "grpclib.server.Stream[GossipRequest, ClusterInfo]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.update(request)
        await stream.send_message(response)

    async def __rpc_read(
        self, stream: "grpclib.server.Stream[_client__.Empty, ClusterInfo]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.read(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.cluster.Gossip/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                GossipRequest,
                ClusterInfo,
            ),
            "/event_store.cluster.Gossip/Read": grpclib.const.Handler(
                self.__rpc_read,
                grpclib.const.Cardinality.UNARY_UNARY,
                _client__.Empty,
                ClusterInfo,
            ),
        }


class ElectionsBase(ServiceBase):
    async def view_change(
        self, view_change_request: "ViewChangeRequest"
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def view_change_proof(
        self, view_change_proof_request: "ViewChangeProofRequest"
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def prepare(self, prepare_request: "PrepareRequest") -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def prepare_ok(
        self, prepare_ok_request: "PrepareOkRequest"
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def proposal(self, proposal_request: "ProposalRequest") -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def accept(self, accept_request: "AcceptRequest") -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def leader_is_resigning(
        self, leader_is_resigning_request: "LeaderIsResigningRequest"
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def leader_is_resigning_ok(
        self, leader_is_resigning_ok_request: "LeaderIsResigningOkRequest"
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_view_change(
        self, stream: "grpclib.server.Stream[ViewChangeRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.view_change(request)
        await stream.send_message(response)

    async def __rpc_view_change_proof(
        self, stream: "grpclib.server.Stream[ViewChangeProofRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.view_change_proof(request)
        await stream.send_message(response)

    async def __rpc_prepare(
        self, stream: "grpclib.server.Stream[PrepareRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.prepare(request)
        await stream.send_message(response)

    async def __rpc_prepare_ok(
        self, stream: "grpclib.server.Stream[PrepareOkRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.prepare_ok(request)
        await stream.send_message(response)

    async def __rpc_proposal(
        self, stream: "grpclib.server.Stream[ProposalRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.proposal(request)
        await stream.send_message(response)

    async def __rpc_accept(
        self, stream: "grpclib.server.Stream[AcceptRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.accept(request)
        await stream.send_message(response)

    async def __rpc_leader_is_resigning(
        self, stream: "grpclib.server.Stream[LeaderIsResigningRequest, _client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.leader_is_resigning(request)
        await stream.send_message(response)

    async def __rpc_leader_is_resigning_ok(
        self,
        stream: "grpclib.server.Stream[LeaderIsResigningOkRequest, _client__.Empty]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.leader_is_resigning_ok(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.cluster.Elections/ViewChange": grpclib.const.Handler(
                self.__rpc_view_change,
                grpclib.const.Cardinality.UNARY_UNARY,
                ViewChangeRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/ViewChangeProof": grpclib.const.Handler(
                self.__rpc_view_change_proof,
                grpclib.const.Cardinality.UNARY_UNARY,
                ViewChangeProofRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/Prepare": grpclib.const.Handler(
                self.__rpc_prepare,
                grpclib.const.Cardinality.UNARY_UNARY,
                PrepareRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/PrepareOk": grpclib.const.Handler(
                self.__rpc_prepare_ok,
                grpclib.const.Cardinality.UNARY_UNARY,
                PrepareOkRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/Proposal": grpclib.const.Handler(
                self.__rpc_proposal,
                grpclib.const.Cardinality.UNARY_UNARY,
                ProposalRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/Accept": grpclib.const.Handler(
                self.__rpc_accept,
                grpclib.const.Cardinality.UNARY_UNARY,
                AcceptRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/LeaderIsResigning": grpclib.const.Handler(
                self.__rpc_leader_is_resigning,
                grpclib.const.Cardinality.UNARY_UNARY,
                LeaderIsResigningRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/LeaderIsResigningOk": grpclib.const.Handler(
                self.__rpc_leader_is_resigning_ok,
                grpclib.const.Cardinality.UNARY_UNARY,
                LeaderIsResigningOkRequest,
                _client__.Empty,
            ),
        }



import asyncio
from typing import AsyncGenerator
from uuid import uuid4

import pytest
from pytest_mock import MockerFixture

from eventstoredb import Client
from eventstoredb.events import JsonEvent
from eventstoredb.streams.subscribe import SubscribeToAllOptions
from eventstoredb.streams.types import (
    AllPosition,
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
    StreamPosition,
)

from ..utils import Subscriber


@pytest.fixture
async def subscriber(
    event_loop: asyncio.AbstractEventLoop,
    mocker: MockerFixture,
) -> AsyncGenerator[Subscriber, None]:
    s = Subscriber(event_loop)
    s.event_handler = mocker.stub()
    yield s
    await s._stop()


async def test_subscribe_to_all(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    subscription = eventstoredb_client.subscribe_to_all()

    subscriber.subscription = subscription

    assert subscriber.event_handler.call_count == 0

    stream_name = f"Test-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="test_subscribe_to_all_Test")],
    )

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]
    events = [e for e in events if e.event.type.startswith("test_subscribe_to_all_")]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_Test"


async def test_subscribe_to_all_from_start(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="test_subscribe_to_all_from_start_Test1")],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(from_position=StreamPosition.START)
    )

    subscriber.subscription = subscription

    assert subscriber.event_handler.call_count == 0

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="test_subscribe_to_all_from_start_Test2")],
    )

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]
    events = [
        e
        for e in events
        if e.event.type.startswith("test_subscribe_to_all_from_start_")
    ]

    assert len(events) == 2

    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "test_subscribe_to_all_from_start_Test1"

    assert events[1].event is not None
    event = events[1].event
    assert event.stream_name == stream_name_2
    assert event.type == "test_subscribe_to_all_from_start_Test2"


async def test_subscribe_to_all_from_end(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="test_subscribe_to_all_from_end_Test1")],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(from_position=StreamPosition.END)
    )

    subscriber.subscription = subscription

    assert subscriber.event_handler.call_count == 0
    await asyncio.sleep(0.1)

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="test_subscribe_to_all_from_end_Test2")],
    )

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]
    events = [
        e for e in events if e.event.type.startswith("test_subscribe_to_all_from_end_")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name_2
    assert event.type == "test_subscribe_to_all_from_end_Test2"


async def test_subscribe_to_all_from_position(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name = f"Test-{uuid4()}"
    # write some events onto stream
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Junk"), JsonEvent(type="Junk")],
    )
    # write one event to get current stream-position
    marker = await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Marker")],
    )
    assert marker.position is not None

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Test")],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            from_position=AllPosition(
                commit=marker.position.commit,
                prepare=marker.position.prepare,
            )
        )
    )

    subscriber.subscription = subscription

    assert subscriber.event_handler.call_count == 0

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]
    events = [e for e in events if not e.event.type.startswith("$")]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "Test"


async def test_subscribe_to_all_exclude_system_events(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(filter=ExcludeSystemEventsFilter())
    )

    subscriber.subscription = subscription

    assert subscriber.event_handler.call_count == 0

    stream_name = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="test_subscribe_to_all_exclude_system_events_Test")],
    )

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]

    events = [
        e
        for e in events
        if e.event.type.startswith("$")
        or e.event.type.startswith("test_subscribe_to_all_exclude_system_events_")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_exclude_system_events_Test"


async def test_subscribe_to_all_filter_by_event_type_regex(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_regex_Test1"),
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_regex_Test2"),
        ],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            filter=EventTypeFilter(
                regex="test_subscribe_to_all_filter_by_event_type_regex_Test1"
            )
        )
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_filter_by_event_type_regex_Test1"


async def test_subscribe_to_all_filter_by_event_type_prefix(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name = f"Test-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_prefix_Test1"),
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_prefix_Test2"),
            JsonEvent(type="something_else"),
        ],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            filter=EventTypeFilter(
                prefix=["test_subscribe_to_all_filter_by_event_type_prefix_"]
            )
        )
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]

    assert len(events) == 2

    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_filter_by_event_type_prefix_Test1"

    assert events[1].event is not None
    event = events[1].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_filter_by_event_type_prefix_Test2"


async def test_subscribe_to_all_filter_by_stream_name_regex(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name_1 = f"test_subscribe_to_all_filter_by_stream_name_regex_Test1-{uuid4()}"
    stream_name_2 = f"test_subscribe_to_all_filter_by_stream_name_regex_Test2-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(filter=StreamNameFilter(regex=stream_name_1))
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]

    assert len(events) == 1
    event = events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "Test"


async def test_subscribe_to_all_filter_by_stream_name_prefix(
    eventstoredb_client: Client,
    subscriber: Subscriber,
) -> None:
    stream_name_1 = (
        f"test_subscribe_to_all_filter_by_stream_name_prefix_Test1-{uuid4()}"
    )
    stream_name_2 = (
        f"test_subscribe_to_all_filter_by_stream_name_prefix_Test2-{uuid4()}"
    )
    stream_name_3 = f"Junk-{uuid4()}"
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_3,
        events=[JsonEvent(type="Test")],
    )

    subscription = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            filter=StreamNameFilter(
                prefix=["test_subscribe_to_all_filter_by_stream_name_prefix_"]
            )
        )
    )

    subscriber.subscription = subscription

    await asyncio.sleep(0.1)
    events = [c[0][0] for c in subscriber.event_handler.call_args_list]

    assert len(events) == 2

    event = events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "Test"

    event = events[1].event
    assert event.stream_name == stream_name_2
    assert event.type == "Test"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: projections.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncIterator, Dict, Optional

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from ... import client as __client__

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class CreateReq(betterproto.Message):
    options: "CreateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptions(betterproto.Message):
    one_time: "__client__.Empty" = betterproto.message_field(1, group="mode")
    transient: "CreateReqOptionsTransient" = betterproto.message_field(2, group="mode")
    continuous: "CreateReqOptionsContinuous" = betterproto.message_field(
        3, group="mode"
    )
    query: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class CreateReqOptionsTransient(betterproto.Message):
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptionsContinuous(betterproto.Message):
    name: str = betterproto.string_field(1)
    emit_enabled: bool = betterproto.bool_field(2)
    track_emitted_streams: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CreateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class UpdateReq(betterproto.Message):
    options: "UpdateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UpdateReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    query: str = betterproto.string_field(2)
    emit_enabled: bool = betterproto.bool_field(3, group="emit_option")
    no_emit_options: "__client__.Empty" = betterproto.message_field(
        4, group="emit_option"
    )


@dataclass(eq=False, repr=False)
class UpdateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DeleteReq(betterproto.Message):
    options: "DeleteReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    delete_emitted_streams: bool = betterproto.bool_field(2)
    delete_state_stream: bool = betterproto.bool_field(3)
    delete_checkpoint_stream: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class DeleteResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StatisticsReq(betterproto.Message):
    options: "StatisticsReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StatisticsReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1, group="mode")
    all: "__client__.Empty" = betterproto.message_field(2, group="mode")
    transient: "__client__.Empty" = betterproto.message_field(3, group="mode")
    continuous: "__client__.Empty" = betterproto.message_field(4, group="mode")
    one_time: "__client__.Empty" = betterproto.message_field(5, group="mode")


@dataclass(eq=False, repr=False)
class StatisticsResp(betterproto.Message):
    details: "StatisticsRespDetails" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StatisticsRespDetails(betterproto.Message):
    core_processing_time: int = betterproto.int64_field(1)
    version: int = betterproto.int64_field(2)
    epoch: int = betterproto.int64_field(3)
    effective_name: str = betterproto.string_field(4)
    writes_in_progress: int = betterproto.int32_field(5)
    reads_in_progress: int = betterproto.int32_field(6)
    partitions_cached: int = betterproto.int32_field(7)
    status: str = betterproto.string_field(8)
    state_reason: str = betterproto.string_field(9)
    name: str = betterproto.string_field(10)
    mode: str = betterproto.string_field(11)
    position: str = betterproto.string_field(12)
    progress: float = betterproto.float_field(13)
    last_checkpoint: str = betterproto.string_field(14)
    events_processed_after_restart: int = betterproto.int64_field(15)
    checkpoint_status: str = betterproto.string_field(16)
    buffered_events: int = betterproto.int64_field(17)
    write_pending_events_before_checkpoint: int = betterproto.int32_field(18)
    write_pending_events_after_checkpoint: int = betterproto.int32_field(19)


@dataclass(eq=False, repr=False)
class StateReq(betterproto.Message):
    options: "StateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StateReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    partition: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class StateResp(betterproto.Message):
    state: "betterproto_lib_google_protobuf.Value" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResultReq(betterproto.Message):
    options: "ResultReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResultReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    partition: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ResultResp(betterproto.Message):
    result: "betterproto_lib_google_protobuf.Value" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResetReq(betterproto.Message):
    options: "ResetReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResetReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    write_checkpoint: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class ResetResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class EnableReq(betterproto.Message):
    options: "EnableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class EnableReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class EnableResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DisableReq(betterproto.Message):
    options: "DisableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DisableReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    write_checkpoint: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class DisableResp(betterproto.Message):
    pass


class ProjectionsStub(betterproto.ServiceStub):
    async def create(
        self,
        create_req: "CreateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "CreateResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Create",
            create_req,
            CreateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def update(
        self,
        update_req: "UpdateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "UpdateResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Update",
            update_req,
            UpdateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete(
        self,
        delete_req: "DeleteReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "DeleteResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Delete",
            delete_req,
            DeleteResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def statistics(
        self,
        statistics_req: "StatisticsReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator["StatisticsResp"]:
        async for response in self._unary_stream(
            "/event_store.client.projections.Projections/Statistics",
            statistics_req,
            StatisticsResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def disable(
        self,
        disable_req: "DisableReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "DisableResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Disable",
            disable_req,
            DisableResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def enable(
        self,
        enable_req: "EnableReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "EnableResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Enable",
            enable_req,
            EnableResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def reset(
        self,
        reset_req: "ResetReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ResetResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Reset",
            reset_req,
            ResetResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def state(
        self,
        state_req: "StateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "StateResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/State",
            state_req,
            StateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def result(
        self,
        result_req: "ResultReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ResultResp":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/Result",
            result_req,
            ResultResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def restart_subsystem(
        self,
        client_empty: "__client__.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "__client__.Empty":
        return await self._unary_unary(
            "/event_store.client.projections.Projections/RestartSubsystem",
            client_empty,
            __client__.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class ProjectionsBase(ServiceBase):
    async def create(self, create_req: "CreateReq") -> "CreateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def update(self, update_req: "UpdateReq") -> "UpdateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self, delete_req: "DeleteReq") -> "DeleteResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def statistics(
        self, statistics_req: "StatisticsReq"
    ) -> AsyncIterator["StatisticsResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def disable(self, disable_req: "DisableReq") -> "DisableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def enable(self, enable_req: "EnableReq") -> "EnableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset(self, reset_req: "ResetReq") -> "ResetResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def state(self, state_req: "StateReq") -> "StateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def result(self, result_req: "ResultReq") -> "ResultResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def restart_subsystem(
        self, client_empty: "__client__.Empty"
    ) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_create(
        self, stream: "grpclib.server.Stream[CreateReq, CreateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.create(request)
        await stream.send_message(response)

    async def __rpc_update(
        self, stream: "grpclib.server.Stream[UpdateReq, UpdateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.update(request)
        await stream.send_message(response)

    async def __rpc_delete(
        self, stream: "grpclib.server.Stream[DeleteReq, DeleteResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete(request)
        await stream.send_message(response)

    async def __rpc_statistics(
        self, stream: "grpclib.server.Stream[StatisticsReq, StatisticsResp]"
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.statistics,
            stream,
            request,
        )

    async def __rpc_disable(
        self, stream: "grpclib.server.Stream[DisableReq, DisableResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.disable(request)
        await stream.send_message(response)

    async def __rpc_enable(
        self, stream: "grpclib.server.Stream[EnableReq, EnableResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.enable(request)
        await stream.send_message(response)

    async def __rpc_reset(
        self, stream: "grpclib.server.Stream[ResetReq, ResetResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.reset(request)
        await stream.send_message(response)

    async def __rpc_state(
        self, stream: "grpclib.server.Stream[StateReq, StateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.state(request)
        await stream.send_message(response)

    async def __rpc_result(
        self, stream: "grpclib.server.Stream[ResultReq, ResultResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.result(request)
        await stream.send_message(response)

    async def __rpc_restart_subsystem(
        self, stream: "grpclib.server.Stream[__client__.Empty, __client__.Empty]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.restart_subsystem(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.projections.Projections/Create": grpclib.const.Handler(
                self.__rpc_create,
                grpclib.const.Cardinality.UNARY_UNARY,
                CreateReq,
                CreateResp,
            ),
            "/event_store.client.projections.Projections/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                UpdateReq,
                UpdateResp,
            ),
            "/event_store.client.projections.Projections/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteReq,
                DeleteResp,
            ),
            "/event_store.client.projections.Projections/Statistics": grpclib.const.Handler(
                self.__rpc_statistics,
                grpclib.const.Cardinality.UNARY_STREAM,
                StatisticsReq,
                StatisticsResp,
            ),
            "/event_store.client.projections.Projections/Disable": grpclib.const.Handler(
                self.__rpc_disable,
                grpclib.const.Cardinality.UNARY_UNARY,
                DisableReq,
                DisableResp,
            ),
            "/event_store.client.projections.Projections/Enable": grpclib.const.Handler(
                self.__rpc_enable,
                grpclib.const.Cardinality.UNARY_UNARY,
                EnableReq,
                EnableResp,
            ),
            "/event_store.client.projections.Projections/Reset": grpclib.const.Handler(
                self.__rpc_reset,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetReq,
                ResetResp,
            ),
            "/event_store.client.projections.Projections/State": grpclib.const.Handler(
                self.__rpc_state,
                grpclib.const.Cardinality.UNARY_UNARY,
                StateReq,
                StateResp,
            ),
            "/event_store.client.projections.Projections/Result": grpclib.const.Handler(
                self.__rpc_result,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResultReq,
                ResultResp,
            ),
            "/event_store.client.projections.Projections/RestartSubsystem": grpclib.const.Handler(
                self.__rpc_restart_subsystem,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
        }



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, auto

from eventstoredb.streams.types import StreamPosition, StreamRevision


class ConsumerStrategy(Enum):
    DISPATCH_TO_SINGLE = auto()
    ROUND_ROBIN = auto()
    PINNED = auto()


@dataclass
class PersistentSubscriptionSettings:
    resolve_links: bool = False
    extra_statistics: bool = False
    message_timeout: int = 30000
    max_retry_count: int = 10
    checkpoint_after: int = 2000
    min_checkpoint_count: int = 10
    max_checkpoint_count: int = 1000
    max_subscriber_count: int = 0
    live_buffer_size: int = 500
    read_batch_size: int = 20
    history_buffer_size: int = 500
    consumer_strategy: ConsumerStrategy = ConsumerStrategy.ROUND_ROBIN


@dataclass
class CreateUpdatePersistentSubscriptionOptions:
    from_revision: StreamPosition | StreamRevision = StreamPosition.START
    settings: PersistentSubscriptionSettings = field(
        default_factory=PersistentSubscriptionSettings
    )



from eventstoredb.streams.read.exceptions import ReadStreamError, StreamNotFoundError
from eventstoredb.streams.read.grpc import (
    convert_read_response,
    create_read_all_request,
    create_read_request,
)
from eventstoredb.streams.read.types import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    ReadAllOptions,
    ReadDirection,
    ReadStreamOptions,
    StreamNameFilter,
)
